// 
// TraceHook - Application responsible for monitoring processes marked
//             as malware together with their children (recursively) and
//             dumping their memory contents after a specified interval or
//             Ctrl+C signal.
// 
// Author:  j00ru//vx
// Date:    2009-08-04
// Version: 0.0.2
// Module:  User-mode client
//
#define _WIN32_WINNT 0x0600
#include <cstdio>
#include <cstdlib>
#include <string>
#include <vector>
#include <windows.h>
#include <ddk\ntapi.h>
#include <psapi.h>
#include <shlwapi.h>
#include "driver.h"
using namespace std;
#define INFO(text) {\
                     printf("[INFO]  ");\
                     printf text;\
                   }
#define DIE(text)  {\
                     printf("[ERROR] ");\
                     printf text;\
                     ExitProcess(1);\
                   }
#define SUCCESS    puts("Success!");



/* The TraceHook device name in a short form
 */
const CHAR ServiceName[]    = "TraceHook";
const CHAR DeviceName[]     = "\\\\.\\TraceHook";
HANDLE hDevice;

/* Kernel-mode module communication routines 
 */
BOOL QueryDriverVersion(HANDLE hDriver);
BOOL SendFinishSignal(HANDLE hDriver);
BOOL AddProcessToList(HANDLE hDriver,DWORD dwProcessId);
BOOL RemoveProcessFromList(HANDLE hDriver,DWORD dwProcessId);

/* Kernel-module loading routines
 */
BOOL LoadDriver();
VOID UnloadDriver();



/* LoadDriver routine
 *
 * Arguments:
 *  None
 * 
 * Return value:
 *  TRUE, if the driver has been successfully loaded and executed
 *  FALSE otherwise
 */
BOOL LoadDriver()
{
  SC_HANDLE hSCManager;
  SC_HANDLE hService;
  SERVICE_STATUS ss;
  CHAR Path[4096];


  /* Before trying to manually load the driver, make sure it
   * is not currently loaded.
   */
  UnloadDriver();

  /* Form the initial version of driver path
   */
  GetCurrentDirectory(sizeof(Path),Path);
  strcat(Path, "\\");
  strcat(Path, DriverFileName);

  /* Open the SC manager
   */
  hSCManager = OpenSCManager(NULL, NULL, SC_MANAGER_CREATE_SERVICE);
  if(!hSCManager)
  {
    DIE(("OpenSCManager() failed\r\n"));
    return false;
  }

  /* Check if the service isn't already active
   */
  hService = OpenService(hSCManager, ServiceName, SERVICE_START | DELETE | SERVICE_STOP | SERVICE_QUERY_STATUS);
  if(!hService)
  {
    /* Try to create our service
     */
    hService = CreateService(hSCManager, ServiceName, ServiceName,
                            	SERVICE_START|DELETE|SERVICE_STOP|SERVICE_QUERY_STATUS, SERVICE_KERNEL_DRIVER,
	                            SERVICE_DEMAND_START, SERVICE_ERROR_IGNORE, Path,
                             	NULL, NULL, NULL, NULL, NULL);
    if(!hService)
    {
      DIE(("CreateService() and OpenService() failed\r\n"));
      CloseServiceHandle(hSCManager);
      return FALSE;
    }
  }

  /* If the service is not already running, try to make it run
   */
  QueryServiceStatus(hService, &ss);
  if(ss.dwCurrentState != SERVICE_RUNNING)
  {
    if(!StartService(hService, 0, NULL))
    {
      DIE(("StartService() failed. LastError: 0x%.8x\r\n",(UINT)GetLastError()));
      ControlService(hService, SERVICE_CONTROL_STOP, &ss);
      DeleteService(hService);
      CloseServiceHandle(hService);
      CloseServiceHandle(hSCManager);
      return FALSE;
    }
  }

  CloseServiceHandle(hService);
  CloseServiceHandle(hSCManager);
  return TRUE;
}


/* UnloadDriver routine
 *
 * Arguments:
 *  None
 * 
 * Return value:
 *  None
 */
VOID UnloadDriver()
{
  SC_HANDLE hSCManager;
  SC_HANDLE hService;
  SERVICE_STATUS ss;

  /* Open the service manager
   */
  hSCManager = OpenSCManager(NULL, NULL, SC_MANAGER_CREATE_SERVICE);

  /* Open the service
   */
  hService = OpenService(hSCManager, ServiceName, SERVICE_START | DELETE | SERVICE_STOP);

  /* Stop and delete the service
   */
  ControlService(hService, SERVICE_CONTROL_STOP, &ss);
  DeleteService(hService);

  /* Close handles
   */
  CloseServiceHandle(hService);
  CloseServiceHandle(hSCManager);
}



/* HandlerRoutine routine
 *
 * Arguments:
 *  None
 * 
 * Return value:
 *  None
 *
 * The function is responsible for signalising the driver that the 
 * client application is being closed, thus forcing it to terminate
 * every process inside the malware list (and therefore dumping them).
 * After that, the driver gets unloaded from the operating system.
 */
void MakeFinalDump()
{
  INFO(("A control event has been detected!\r\n"));
  INFO(("Sending a FINISH signal to the driver...\r\n"));
  RemoveProcessFromList(hDevice,GetCurrentProcessId());
  SendFinishSignal(hDevice);

  INFO(("Unloading the driver from operating system.\r\n"));
  UnloadDriver();
}


/* QueryDriverVersion routine
 *
 * Arguments:
 *  hDevice - handle to the active TraceHook driver
 * 
 * Return value:
 *  TRUE, if the version has been successfully received
 *  FALSE otherwise
 */
BOOL QueryDriverVersion(HANDLE hDriver)
{
  static char VersionBuffer[4096];
  DWORD BytesReturned;
  
  /* Send an IOCTL signal to the driver and wait for a response
   */
  if(!DeviceIoControl(hDriver,
                      IOCTL_QUERY_VERSION,
                      NULL,
                      0,
                      VersionBuffer,
                      sizeof(VersionBuffer),
                      &BytesReturned,
                      NULL))
  {
    DIE(("DeviceIoControl() failed. LastError: 0x%.8x\n",(UINT)GetLastError()));
    return FALSE;
  }

  /* Print the obtained string
   */
  INFO(("Driver Version: %s\n",VersionBuffer));
  return TRUE;
}


BOOL SendFinishSignal(HANDLE hDriver)
{
  DWORD BytesReturned;

  /* Send an IOCTL signal to the driver and wait for a response
   */
  if(!DeviceIoControl(hDriver,
                      IOCTL_FINISH,
                      NULL,
                      0,
                      NULL,
                      0,
                      &BytesReturned,
                      NULL))
  {
    DIE(("DeviceIoControl() failed. LastError: 0x%.8x\n",(UINT)GetLastError()));
    return FALSE;
  }

  return TRUE;  
}


/* AddProcessToList routine
 *
 * Arguments:
 *  dwProcessId - PID of the process to add to the tracing list
 * 
 * Return value:
 *  TRUE, if the signal has been successfully handled by the driver
 *  FALSE otherwise
 */
BOOL AddProcessToList(HANDLE hDriver,DWORD dwProcessId)
{
  DWORD BytesReturned;

  /* Send an IOCTL signal to the driver and wait for a response
   */
  if(!DeviceIoControl(hDriver,
                      IOCTL_ADD_PROCESS,
                      &dwProcessId,
                      sizeof(DWORD),
                      NULL,
                      0,
                      &BytesReturned,
                      NULL))
  {
    DIE(("DeviceIoControl() failed. LastError: 0x%.8x\n",(UINT)GetLastError()));
    return FALSE;
  }

  return TRUE;
}


/* RemoveProcessFromList routine
 *
 * Arguments:
 *  dwProcessId - PID of the process to remove from the tracing list
 * 
 * Return value:
 *  TRUE, if the signal has been successfully handled by the driver
 *  FALSE otherwise
 */
BOOL RemoveProcessFromList(HANDLE hDriver,DWORD dwProcessId)
{
  DWORD BytesReturned;

  /* Send an IOCTL signal to the driver and wait for a response
   */
  if(!DeviceIoControl(hDriver,
                      IOCTL_REMOVE_PROCESS,
                      &dwProcessId,
                      sizeof(DWORD),
                      NULL,
                      0,
                      &BytesReturned,
                      NULL))
  {
    DIE(("DeviceIoControl() failed. LastError: 0x%.8x\n",(UINT)GetLastError()));
    return FALSE;
  }

  return TRUE;
}



/* main routine
 *
 * Arguments:
 *  Not important (information about the console arguments)
 * 
 * Return value:
 *  Not important
 */
int main(int argc, char** argv)
{
  puts("-=*( TraceHook v0.0.1 by j00ru//vx )*=-");

  if(argc!=2)
  {
    DIE(("Not enough arguments!\n"
        " Usage: TraceHook.exe <target executable>\n"));
    return 1;
  }

  INFO(("Loading driver... "));
  if(!LoadDriver())
    DIE(("Failed"));
  SUCCESS;

  INFO(("Opening driver... "));
  hDevice = CreateFileA(DeviceName,GENERIC_READ|GENERIC_WRITE,0,NULL,OPEN_EXISTING,0,NULL);
  if(hDevice==NULL)
  {
    UnloadDriver();
    DIE(("Failed"));
  }
  SUCCESS;  

  INFO(("Querying for the driver version...\r\n"));
  if(!QueryDriverVersion(hDevice))
  {
    UnloadDriver();
    DIE(("Failed"));
  }

  INFO(("Adding the current ProcessId to the monitor list.\r\n"));
  AddProcessToList(hDevice,GetCurrentProcessId());

  INFO(("Creating the %s directory for dump files.\r\n",UserDumpPath));
  CreateDirectoryA(UserDumpPath,NULL);


  /* Create the specified process
   */
  STARTUPINFO si;
  PROCESS_INFORMATION pi;
  memset(&si,0,sizeof(si));
  memset(&pi,0,sizeof(pi));
  si.cb = sizeof(si);

  INFO(("Creating the target process..."));
  if(!CreateProcess(NULL,argv[1],NULL,NULL,false,CREATE_NEW_CONSOLE,NULL,NULL,&si,&pi))
  {
    RemoveProcessFromList(hDevice,GetCurrentProcessId());
    UnloadDriver();
    DIE(("Failed"));
  }
  SUCCESS;

  INFO(("TraceHook module is active, monitoring the malware processes.\r\n\r\n"));
  INFO(("Press ENTER to stop the analysis.\r\n"));

  /* Wait for the user to press ENTER
   */
  getchar();

  /* Inform the driver about the process termination, so as to let it 
   * dump all the remaining processes - after it is done, the driver
   * gets unloaded
   */
  MakeFinalDump();
  return 0;
}

